<?php
/**
 * @file
 * Functions for importing releases from Discogs.com.
 */

/**
 * Base Discogs.com URL.
 */
define('DISCOGS_BASE_URL', 'https://api.discogs.com/');

/**
 * Form for importing releases from Discogs.com into Drupal. This is the first
 * form stage, where the user enters a keyword and type to fetch a list of
 * releases (or, if that fails, a list of artists/labels to resubmit).
 * Those releases will be displayed in subsequent form stages, generated by
 * discogs_import_releases_form().
 *
 * @param $form_state
 *      Form state
 * @return
 *      Multidimensional $form array
 * @ingroup forms
 * @see discogs_import_releases_form()
 */
function discogs_import_form(&$form_state) {
  // Display results
  if (isset($form_state['storage']['step'])
      && ($form_state['storage']['step'] > 0)
      && is_array($form_state['storage']['data'])) {
    return discogs_import_releases_form($form_state);
  }
  $form_state['storage']['step'] = 0;

  $form['term'] = array(
    '#type' => 'textfield',
    '#title' => t('Release, Artist, Label, or Master Release'),
    '#description' => t('Release, Artist, Label, or Master Release to pull from Discogs.com. If you know the exact ID, artist name, or label name, you will be presented with all the relevant releases to import. If not, a search will be performed, and you may then choose which releases you want to import. If your search turns up unexpected results, try searching with a more specific phrase.'),
  );
  $form['type'] = array(
    '#type' => 'select',
    '#default_value' => 'release',
    '#options' => array(
      'release' => t('Release'),
      'artist' => t('Artist'),
      'label' => t('Label'),
      'master' => t('Master Release'),
    ),
  );
  // Custom submit
  $form['next'] = array(
        '#type' => 'submit',
        '#value' => 'Next >>',
  );

  return $form;
}

/**
 * This function generates the subsequent form stages for importing releases
 * from Discogs.com. At this point, the user has either returned an array of
 * releases, or an array of data to resubmit. In either case, it will be
 * presented to the user as a checkbox. The form submit function will work
 * out whether to submit it, or re-display it with fetched releases.
 *
 * @param $form_state
 *      Form state
 * @return
 *      Multidimensional $form array
 * @ingroup forms
 * @see discogs_import_form()
 */
function discogs_import_releases_form($form_state) {
  // Informational form element
  if (isset($form_state['storage']['info'])
      && !empty($form_state['storage']['info'])) {
    $form['info'] = array(
      '#value' => theme_discogs_forminfo($form_state['storage']['info']),
    );
  }

  // Format releases array for checkbox
  $chex = array();

  foreach ($form_state['storage']['data'] as $did => $data) {
    $text = theme('discogs_checkbox_text', $data);
    $id = isset($data['id']) ? $data['id'] : $did;
    $chex[$id] = $text;
  }
  // Build check boxes
  $btn_type = ($form_state['storage']['step'] == 1) ? 'radios' : 'checkboxes';
  $form['data'] = array(
    '#type' => $btn_type,
    '#title' => t('Returned'),
    '#options' => $chex,
  );
  // Set the #value of our submit button, depending upon the import stage
  switch ($form_state['storage']['step']) {
    case 2:
      $value = t('Import');
      break;
    case 1:
      $value = t('Get Releases');
      break;
    default:
      $value = t('Next >>');
  }
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => $value,
  );
  return $form;
}

/**
 * Submit handler for discogs_import_form().
 *
 * Will decide at what step of the form process we are at, and set form
 * variables accordingly:
 * 0 = First submitted
 * 1 = Searched, returned list of labels or artists (not releases)
 * 2 = Final import of releases
 *
 * @param $form
 *   Form array
 * @param $form_state
 *   Form state
 * @return
 */
function discogs_import_form_submit($form, &$form_state) {
  if (!isset($form_state['storage']['step'])) {
    $form_state['storage']['step'] = 0;
  }
  switch ($form_state['storage']['step']) {
    case 1:
      _discogs_form_step_1($form, $form_state);
      break;
    case 2:
      _discogs_form_step_2($form, $form_state);
      break;
    default:
      _discogs_form_step_0($form, $form_state);
  }
}

/**
 * Helper function for discogs_import_form_sumbit().
 *
 * Handles the initial form state: when the form is first submitted.
 *
 * @param $form
 *   Form array
 * @param $form_state
 *   Form state
 * @return
 */
function _discogs_form_step_0($form, &$form_state) {
  $results = array();
  $step = $form_state['storage']['step'];
  $term = $form_state['values']['term'];
  $type = $form_state['values']['type'];

  // If search term is numeric, assume it's an ID
  if (is_numeric($term)) {
    $url = DISCOGS_BASE_URL . $type . 's/' . $term;
    switch ($type) {
      case 'release':// Release ID
        $qr = _discogs_http_request($url);
        if (!empty($qr)) {
          $results['results'][0] = $qr;
          $results['pagination'] = array(
            'per_page' => 50,
            'items' => 1,
            'page' => 1,
          );
        }
        break;
      case 'label':  // Deliberate fallthrough
      case 'artist': // Label or artist ID
        $url .= '/releases';
        $qr = _discogs_http_request($url);
        if (!empty($qr)) {
          $results['results'] = $qr['releases'];
          $results['pagination'] = $qr['pagination'];
        }
        break;
      case 'master': // Master recording ID
        $url .= '/versions';
        $qr = _discogs_http_request($url);
        if (!empty($qr)) {
          $results['results'] = $qr['versions'];
          $results['pagination'] = $qr['pagination'];
        }
        break;
    }
    $step = 2;
  }
  if (empty($results)) {
    // Not ID, or ID returned no results
    $url = DISCOGS_BASE_URL . "database/search";
    $results = _discogs_http_request($url, array('q' => $term, 'type' => $type));
    $step = ($type == 'release') ? 2 : 1;
  }

  if ($results['results']) {
    // Store all the search result info
    $form_state['storage']['data'] = $results['results'];
    $form_state['storage']['pagination'] = $results['pagination'];
    $form_state['storage']['step'] = $step;
    $form_state['storage']['type'] = $type;
  }
  else {
    // Search returned nothing at all
    drupal_set_message(t('No releases found.'));
    unset($form_state['storage']);
    $form_state['storage']['step'] = 0;
  }
}

/**
 * Helper function for discogs_import_form_sumbit().
 *
 * Handles the first form state: the user did a search for releases by
 * artist or label.
 *
 * @param $form
 *   Form array
 * @param $form_state
 *   Form state
 * @return
 */
function _discogs_form_step_1($form, &$form_state) {
  // Get submitted value
  $value = $form_state['values']['data'];
  $data  = $form_state['storage']['data'];

  // Put the values into the form data, and re-do the form submission
  $form_state['values']['type']  = $form_state['storage']['type'];
  $form_state['values']['term']  = $value;
  $form_state['storage']['step'] = 0;
  _discogs_form_step_0($form, $form_state);
}

/**
 * Helper function for discogs_import_form_sumbit().
 *
 * Handles the third (and final) form state: the user selected which releases
 * to import, and now those must be turned into nodes.
 *
 * @param $form
 *   Form array
 * @param $form_state
 *   Form state
 * @return
 */
function _discogs_form_step_2($form, &$form_state) {
  module_load_include('inc', 'node', 'node.pages');
  $releases = $form_state['values']['data'];
  $releases = empty($releases) ? array() : array_values($releases);
  foreach ($releases as $release) {
    if (!empty($release) && ($release > 0)) {
      // Handle master releases
      if($form_state['storage']['data'][$release]['type'] == 'master') {
        $release = $form_state['storage']['data'][$release]['main_release'];
      }
      $url = DISCOGS_BASE_URL . "releases/$release";
      $query_result = _discogs_http_request($url);

      if ($query_result) {
        // Create form state from Discogs data
        $form_state = array();
        $form_state['values'] = discogs_parse_release($query_result);
        $form_state['track_count'] = count($form_state['values']['tracks']);
        // Create a new node
        $node = new stdClass();
        $node->type = 'discogs';
        // For some reason, I have to insert files into the node directly
        $node->files = _discogs_get_filenodes($form_state);
        // Fixes [#1428150]: Import data being stored with uid of 0
        global $user;
        $node->name = $user->name;
        drupal_execute('discogs_node_form', $form_state, $node);
      }
      else {
        drupal_set_message(t('Error importing from Discogs.com'));
      }
    }
  }
  // Redirect to content admin page
  unset($form_state['storage']);
  $form_state['redirect'] = 'node/add/discogs_import';
}

/**
 * Searches Discogs and returns an array of the received JSON data.
 *
 * @param $url
 *     URL of discogs term
 * @return
 *     Multi-dimensional array containing all releases
 */
function _discogs_http_request($base, $query_data = array()) {
  $json_array = array();

  // Discogs Auth info
  $consumer_key = variable_get('discogs_key', '');
  $consumer_secret = variable_get('discogs_secret', '');
  if (empty($consumer_key) || empty($consumer_secret)) {
    // Can't search without OAuth keys...
    $msg = t('You must provide a Consumer Key and Consumer Secret.');
    $msg .= t('Go to the !link page to enter these values.',
        array('!link' => l(t('Discogs Settings'), 'admin/settings/discogs')));
    drupal_set_message($msg, 'error');
    return $json_array;
  }

  // Build the fully-qualified URL
  $url = $base;
  if (!empty($query_data)) {
    $url .= '?' . http_build_query($query_data);
  }
  $headers = array(
    // Discogs.com requires a descriptive user-agent
    'User-Agent' => 'DiscogsForDrupal/6.x-1.x +http://drupal.org/project/discogs',
    // Apparently, you have to authorize using the headers
    'Authorization' => "Discogs key=$consumer_key, secret=$consumer_secret"
  );
  $req = drupal_http_request($url, $headers);
  if ($req->code == "200") {
    $json_array = json_decode($req->data, TRUE);
  }
  else {
    // Hopefully this should never happen
    $msg = $url . ' ' . t('returned') . ' ' . $req->code . ': ' . $req->error;
    drupal_set_message($msg, 'error');
  }
  return $json_array;
}

/**
 * Parses a single release into an array that is suitable to be passed to the
 * discogs_form() form handler as $form_state. We will also need to do
 * validation here, as it will not be done when the node is created.
 *
 * @param $releases
 *      A single release from Discogs.com - Important! NOT an array of releases!
 * @param $cleanup
 *      Whether or not to clean up artist fields (default TRUE)
 * @return
 *      Multi-dimensional array to pass to discogs_form()
 * @see _discogs_cleanup_artist_field()
 * @see discogs_form()
 */
function discogs_parse_release($release, $cleanup = TRUE) {
  $new_form = array();

  // Theme artist field (in case of split releases)
  $discog_artist = theme('discogs_artists', $release['artists'], $cleanup);
  // Theme label field (in case of multiple labels)
  $discog_labels = theme('discogs_labels', $release['labels']);
  // Theme format array
  $discog_format = theme('discogs_formats', $release['formats']);
  // Theme release's 'extraartists' as 'discog_credits'
  $discog_credits = theme('discogs_extra_artists', $release['extraartists'],
      $cleanup);

  // Get album info from $release array
  $new_form['title']          = $release['title'];
  $new_form['did']            = $release['id'];
  $new_form['discog_artist']  = $discog_artist;
  $new_form['discog_label']   = $discog_labels['discog_label'];
  $new_form['discog_cat_num'] = $discog_labels['discog_cat_num'];
  $new_form['discog_format']  = $discog_format;
  $new_form['discog_date']    = $release['released'];
  $new_form['discog_country'] = $release['country'];
  $new_form['discog_notes']   = $release['notes'];
  $new_form['discog_credits'] = $discog_credits;

  // Get array of tracks
  $tracks = $release['tracklist'];
  foreach ($tracks as $track) {
    $track_position = $track['position'];
    // Check if $position is empty, because sometimes Discogs "tracks"
    // aren't really tracks
    if (empty($track_position)) {
      unset($track);
      continue;
    }
    $track_artist = is_array($track['artists'])
        ? theme('discogs_artists', $track['artists']['artist'], $cleanup) : '';
    // Need to make sure title isn't more that 128 characters long
    $track_title = _discogs_trim($track['title']);
    // Theme track's 'extraartists' as 'track_notes'
    $track_notes = theme('discogs_extra_artists', $track['extraartists'],
        $cleanup);
    // For some reason, tracks without a duration return an empty array
    $track_duration = is_array($track['duration']) ? '' : $track['duration'];
    $tracks_array[] = array(
      'track_position' => $track_position,
      'track_artist'   => $track_artist,
      'track_title'    => $track_title,
      'track_duration' => $track_duration,
      'track_notes'    => $track_notes,
      'track_lyrics'   => ''
      );
  }
  $new_form['tracks'] = $tracks_array;

  // Handle release image
  $img_fields = _discogs_handle_images($release);

  $new_form = $new_form + $img_fields;
  //*/

  $new_form['op'] = t('Save');
  return $new_form;
}

/**
 * Handles the images associated with a release.
 *
 * @param $release
 *      Array containing all information for a single release
 * @param $title
 *      Title of image node (used by Image module)
 * @return
 *      Array to insert into $form_state['values']
 * @see _discogs_copy_remote_file()
 * @see _discogs_copy_to_files()
 * @see _discogs_copy_to_images()
 */
function _discogs_handle_images($release) {
  $form        = array();
  $urls        = array();
  $has_img     = isset($release['images']);
  $img_handler = variable_get('discogs_img_handler', 0);
  if ($has_img) {
    $urls = $release['images'];
  }

  // Pass images to CCK imagefield module, if applicable
  if (module_exists('imagefield') && ($img_handler == 2) && $has_img) {
    $form['field_discogs_cover'] = _discogs_copy_to_files($urls);
  }

  // Pass images to image_attach, if applicable
  elseif (module_exists('image_attach') && ($img_handler == 3)) {
    $form['iids'] = $has_img
        ? _discogs_copy_to_images($urls, $release['title']) : array();
  }

  // Pass images to core upload module, if applicable
  elseif (module_exists('upload') && ($img_handler == 1) && $has_img) {
    $copied_files = _discogs_copy_to_files($urls);
    // The files field expects an array with fid as key
    foreach ($copied_files as $file) {
      $files[$file['fid']] = $file + array(
        'description' => $file['filename'],
        'remove' => FALSE,
        'list' => 1,
        'new' => TRUE,
        'weight' => 0,
      );
    }
    $form['files'] = $files;
  }

  // If nothing else is available, just store the URL's in the node
  elseif ($has_img) {
    $discog_img            = _discogs_copy_remote_file($urls[0]['uri']);
    $discog_img150         = _discogs_copy_remote_file($urls[0]['uri150']);
    $form['discog_img']    = $discog_img ? $discog_img : '' ;
    $form['discog_img150'] = $discog_img150 ? $discog_img150 : '' ;
  }

  return $form;
}

/**
 * Copies a binary file from a remote URL to the sub-directory of
 * file_directory_path() set by the discogs_img_dir variable.
 *
 * @param $url
 *      URL of file to copy
 * @return
 *      URL of current filename (if copying is successful), or FALSE (if not)
 * @see _discogs_copy_remote_file(
 */
function _discogs_copy_remote_file($url) {
  if (empty($url)) {
    return FALSE;
  }
  $obj = drupal_http_request($url);
  if (!$obj->data) {
    $error = t("Failed to open %url", array('%url' => $url));
    drupal_set_message($error, 'error');
    return FALSE;
  }
  else {
    $discogs_img_dir = variable_get('discogs_img_dir', '');
    $files_subdir = empty($discogs_img_dir)
        ? '' : rtrim($discogs_img_dir, '/\\') . '/';
    $to = file_directory_path() . '/' . $files_subdir . basename($url);
    @$fc = fopen($to, "wb");
    if (!$fc) {
      $error = t("Could not write to %to", array('%to' => $to));
      drupal_set_message($error, 'error');
      return FALSE;
    }
    fwrite($fc, $obj->data);
    fclose($fc);
    return $to;
  }
}

/**
 * Takes an array of URL's, copies them to the files directory, and returns
 * an array suitable for inserting into $form_state.
 *
 * @param $urls
 *   Array of URL's to import
 * @return
 *   Array to insert into $form_state
 * @see _discogs_handle_images()
 * @see _discogs_copy_remote_file()
 */
function _discogs_copy_to_files($urls) {
  $i     = 0;
  $files = array();
  global $user;
  foreach ($urls as $url) {
    $img_url = _discogs_copy_remote_file($url['uri']);
    if ($img_url) {
      // Create file object and update files table
      $file            = new stdClass();
      $file->uid       = $user->uid;
      $file->filename  = basename($img_url);
      $file->filepath  = $img_url;
      $file->filemime  = file_get_mimetype(basename($img_url));
      $file->filesize  = filesize($img_url);
      $file->status    = 0;
      $file->timestamp = time();
      drupal_write_record('files', $file);

      // Create array for form submission
      $files[$i]            = (array) $file;
      $files[$i]['_weight'] = $i;
      $i++;
    }
  }
  return $files;
}

/**
 * Helper function to create image nodes, and return an array of their nid's
 * to put into $form_state['values']['iids']
 *
 * @param $urls
 *   Array of (local) URL's to images
 * @param $title
 *   Title of the release
 * @return
 *   Array of image node nid's
 */
function _discogs_copy_to_images($urls, $title = '') {
  // Need an iterator to make sure we don't get too many images
  $i   = 1;
  $max = variable_get('image_attach_maximum_discogs', 0);
  foreach ($urls as $url) {
    if (($i > $max) && ($max > 0)) {
      continue;
    }
    $img_url = _discogs_copy_remote_file($url['uri']);
    if ($img_url) {
      $new_title = empty($title)
          ? t('Release image @i', array('@i' => $i))
          : t('@title image @i', array('@title' => $title, '@i' => $i));
      $node = image_create_node_from($img_url, $new_title);
      if ($node) {
        $form[] = $node->nid;
      }
      else {
        $error = t("Could not create image node from %url",
        array('%url' => $img_url));
        drupal_set_message($error, 'error');
      }
    }
    $i++;
  }
  return $form;
}

/**
 * Takes $form_state, and returns an array of file node objects to be inserted
 * into $node->files.
 *
 * @param $form_state
 *   Drupal form state
 * @return
 *   Array of file node objects
 */
function _discogs_get_filenodes($form_state) {
  $filenodes = array();
  if (is_array($form_state['values']['files'])) {
    foreach ($form_state['values']['files'] as $file) {
      $filenodes[] = (object) $file;
    }
  }
  return $filenodes;
}